//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      GUI/View/SampleDesigner/SampleListView.cpp
//! @brief     Implements class SampleListView
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2021
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#include "GUI/View/SampleDesigner/SampleListView.h"
#include "GUI/Application/ApplicationSettings.h"
#include "GUI/Model/Sample/GUIExamplesFactory.h"
#include "GUI/Model/Sample/MultiLayerItem.h"
#include "GUI/Model/Sample/SampleListModel.h"
#include "GUI/Model/State/SessionData.h"
#include "GUI/View/Common/ItemViewOverlayButtons.h"
#include "GUI/View/Tool/ItemDelegateForHTML.h"
#include <QAction>
#include <QDialog>
#include <QDialogButtonBox>
#include <QFormLayout>
#include <QLineEdit>
#include <QMenu>
#include <QPainter>
#include <QPushButton>
#include <QTextEdit>

namespace {

class ItemDelegateForSampleTree : public ItemDelegateForHTML {
public:
    ItemDelegateForSampleTree(QObject* parent) : ItemDelegateForHTML(parent) {}

protected:
    void paint(QPainter* painter, const QStyleOptionViewItem& option,
               const QModelIndex& index) const
    {
        ItemDelegateForHTML::paint(painter, option, index);

        QStyleOptionViewItem options = option;
        initStyleOption(&options, index);

        painter->save();

        painter->setPen(QPen(Qt::lightGray, 1));
        painter->drawLine(options.rect.left(), options.rect.bottom(), options.rect.right(),
                          options.rect.bottom());
        painter->restore();
    }

    QSize sizeHint(const QStyleOptionViewItem& option, const QModelIndex& index) const
    {
        auto s = ItemDelegateForHTML::sizeHint(option, index);
        s.setHeight(std::max(s.height(), 32));
        return s;
    }
};

} // namespace

SampleListView::SampleListView(QWidget* parent, SampleModel* sampleModel) : QListView(parent)
{
    m_model = new SampleListModel(this, sampleModel);

    setContextMenuPolicy(Qt::CustomContextMenu);
    setModel(m_model);
    setVerticalScrollMode(QAbstractItemView::ScrollPerPixel);

    ItemViewOverlayButtons::install(
        this, [=](const QModelIndex& i, bool h) { return getOverlayActions(i, h); });
    setItemDelegate(new ItemDelegateForSampleTree(this));

    connect(selectionModel(), &QItemSelectionModel::currentChanged, this,
            &SampleListView::onCurrentChanged);

    connect(this, &QWidget::customContextMenuRequested, this, &SampleListView::showContextMenu);

    m_newSampleAction = new QAction(this);
    m_newSampleAction->setText("Create new sample");
    m_newSampleAction->setIcon(QIcon(":/images/shape-square-plus.svg"));
    m_newSampleAction->setIconText("New");
    m_newSampleAction->setToolTip("Create new sample");
    connect(m_newSampleAction, &QAction::triggered, this, &SampleListView::createNewSample);

    m_chooseFromLibraryAction = new QAction(this);
    m_chooseFromLibraryAction->setText("Choose from sample library");
    m_chooseFromLibraryAction->setIcon(QIcon(":/images/library.svg"));
    m_chooseFromLibraryAction->setIconText("Library");
    m_chooseFromLibraryAction->setToolTip("Choose from sample library");

    QMenu* menu = new QMenu(this);
    m_chooseFromLibraryAction->setMenu(menu);

    for (auto exampleName : GUIExamplesFactory::exampleNames()) {
        QString title, description;
        std::tie(title, description) = GUIExamplesFactory::exampleInfo(exampleName);
        auto icon = QIcon(":/SampleDesignerToolbox/images/sample_layers2.png");
        auto* action = menu->addAction(icon, title);
        action->setToolTip(description);
        connect(action, &QAction::triggered,
                [=]() { createSampleFromLibrary(exampleName, title, description); });
    }
}

void SampleListView::setCurrentSample(MultiLayerItem* multiLayer)
{
    setCurrentIndex(m_model->indexForItem(multiLayer));
}

MultiLayerItem* SampleListView::currentSample()
{
    return m_model->itemForIndex(currentIndex());
}

QAction* SampleListView::newSampleAction()
{
    return m_newSampleAction;
}

QAction* SampleListView::chooseFromLibraryAction()
{
    return m_chooseFromLibraryAction;
}

QSize SampleListView::sizeHint() const
{
    QSize s = QListView::sizeHint();
    s.setWidth(std::max(300, s.width()));
    return s;
}

void SampleListView::createNewSample()
{
    const QModelIndex newIndex = m_model->createSample();
    setCurrentIndex(newIndex);
}

void SampleListView::createSampleFromLibrary(const QString& classname, const QString& title,
                                             const QString& description)
{
    const QModelIndex newIndex = m_model->createSampleFromExamples(
        classname, title, description, gSessionData->projectDocument->materialModel());
    setCurrentIndex(newIndex);
}

QList<QAction*> SampleListView::getOverlayActions(const QModelIndex& index, bool asHover)
{
    if (!asHover)
        return {};

    auto item = m_model->itemForIndex(index);
    if (item == nullptr)
        return {};

    return {createEditAction(this, item), createRemoveAction(this, item)};
}

void SampleListView::editNameAndDescription(MultiLayerItem* item)
{
    QDialog dlg(this);
    dlg.setObjectName("NameAndDescriptionDialog");
    dlg.setWindowTitle("Edit name and description of sample");
    dlg.setWindowFlag(Qt::WindowContextHelpButtonHint, false);
    dlg.setProperty("stylable", true); // for stylesheet addressing

    QFormLayout* layout = new QFormLayout(&dlg);
    QLineEdit* nameEdit = new QLineEdit(&dlg);
    nameEdit->setText(item->itemName());
    nameEdit->selectAll();

    QTextEdit* descriptionEdit = new QTextEdit(&dlg);
    descriptionEdit->setMinimumWidth(200);
    descriptionEdit->setAcceptRichText(false);
    descriptionEdit->setTabChangesFocus(true);
    descriptionEdit->setPlainText(item->description());

    QDialogButtonBox* buttonBox = new QDialogButtonBox(&dlg);
    buttonBox->addButton(QDialogButtonBox::Ok);
    buttonBox->addButton(QDialogButtonBox::Cancel);
    buttonBox->button(QDialogButtonBox::Ok)->setAutoDefault(true);
    buttonBox->button(QDialogButtonBox::Ok)->setDefault(true);

    layout->addRow("Name:", nameEdit);
    layout->addRow("Description:", descriptionEdit);
    layout->addWidget(buttonBox);
    dlg.setLayout(layout);

    connect(buttonBox, &QDialogButtonBox::accepted, &dlg, &QDialog::accept);
    connect(buttonBox, &QDialogButtonBox::rejected, &dlg, &QDialog::reject);

    appSettings->loadWindowSizeAndPos(&dlg);
    if (dlg.exec() == QDialog::Accepted) {
        QModelIndex index = m_model->indexForItem(item);
        m_model->setData(index, nameEdit->text(), Qt::EditRole);
        m_model->setData(index, descriptionEdit->toPlainText(), Qt::ToolTipRole);
        // if length of description changes height of TreeItem, the position of the overlay buttons
        // has to be updated -> schedule a re-layout
        scheduleDelayedItemsLayout();
    }
    appSettings->saveWindowSizeAndPos(&dlg);
}

void SampleListView::onCurrentChanged(const QModelIndex& index)
{
    emit currentSampleChanged(m_model->itemForIndex(index));
}

QAction* SampleListView::createEditAction(QObject* parent, MultiLayerItem* item)
{
    QAction* editAction = new QAction(parent);
    editAction->setText("Edit name and description");
    editAction->setIcon(QIcon(":/images/edit.svg"));
    editAction->setIconText("Edit");
    editAction->setToolTip("Edit name and description");
    connect(editAction, &QAction::triggered, [=]() { editNameAndDescription(item); });
    return editAction;
}


QAction* SampleListView::createRemoveAction(QObject* parent, MultiLayerItem* item)
{
    QAction* removeAction = new QAction(parent);
    removeAction->setText("Remove");
    removeAction->setIcon(QIcon(":/images/delete.svg"));
    removeAction->setIconText("Remove");
    removeAction->setToolTip("Remove this sample");
    connect(removeAction, &QAction::triggered, [=]() { m_model->removeSample(item); });

    return removeAction;
}

void SampleListView::showContextMenu(const QPoint& pos)
{
    auto sampleAtPoint = m_model->itemForIndex(indexAt(pos));

    QMenu menu;
    menu.setToolTipsVisible(true);

    if (sampleAtPoint != nullptr) {
        menu.addAction(createEditAction(&menu, sampleAtPoint));
        menu.addAction(createRemoveAction(&menu, sampleAtPoint));
        menu.addSeparator();
    }

    menu.addAction(m_newSampleAction);
    menu.addAction(m_chooseFromLibraryAction);
    menu.exec(mapToGlobal(pos));
}
